function soard(tr,te,first)

%function soard(tr,te,first) implements the Self-Organizing ARTMAP rule
%discovery system. Input data used during Stage 1 is received in tr, input
%data for Stage 2 is in te, and first allows to use a new order of input
%vectors (first=1) or one previously saved (first=0).

global mato sigma_k wrltemp nwrltemp div

%Initializing variables...
ncat=6;                    %Number of classes
voters=1;                  %Number of voting systems
rob=0.0;                   %Base vigilance
alfa=0.01;                 %Signal rule parameter
epsilon=-0.001;            %Match tracking parameter
p=1;                       %CAM power rule
Bf=1;                      %Learning fraction 
cap=1000000;               %Maximum number of vectors per label during Stage 1

exper=zeros(1,ncat);        %Experience (Actual count)
experO=exper;               %Experience old (Continuously increasing)
xp=exper;
maxXP=6*ncat;               %Maximum experience

classcountH=zeros(1,ncat,voters);   %Counts the number of nodes in each category
timeslearned=zeros(1,ncat,voters);  %Counts the number of times an input vector of a specific class is processed
ok=0;                               %Indicates if an appropriate node was found in WTA mode
tiecount=0;
contamo=0;
contamo2=0;

predictoH=zeros(1,ncat);
predatoH=zeros(1,ncat);

[m,n]=size(tr);             %m is the number of vectors presented in Stage 1
tr(:,1)=tr(:,1)+1;          %Assuming labels start at 0
M=n-1;                      %Number of input features
Tu=M;

%Weights initialization
wijH=ones(2*M,1,voters);    %For the Hybrid version
wjkH=zeros(1,ncat,voters);
wjiH=ones(2*M,1,voters);

wfab=zeros(ncat,ncat);      %Weights in layer Fab
mato=wfab;
wrl=zeros(ncat,ncat);       %Initializing numerator for Wfab self-organization
wrltemp=wrl;
nwrltemp=wrl;
div=wrl;

%ordenes generates voters different data orderings to use during Stage 1
train=ordenes(tr,voters);   

%Complement coding
comp=[train 1-train];       
A=[comp(:,1:n,:) comp(:,n+2:2*n,:)];

%Save new ordering or load an old one
if first == 1
    save trainwithis1 A;
else
    load trainwithis1;
end;

%Stage 1

%Repeat for all voters
for vote=1:voters,
    
    TdistH=zeros(1,2);       %(:,1) holds the value and (:,2) the corresponding node number
    
    %For the first input... the first committed node
    C=1;
    wijH(:,C,vote)=A(1,2:2*n-1,vote)';  %For the first input vector
    wjiH(:,C,vote)=wijH;
    taoijH(:,C,vote)=1-wijH(:,C,vote);
    taojiH(:,C,vote)=1-wjiH(:,C,vote);  %Initializing feedback weights
    wjkH(C,A(1,1,vote),vote)=1;
    classcountH(1,A(1,1,vote),vote)=classcountH(1,A(1,1,vote))+1;
    timeslearned(1,:,vote)=classcountH;
    
    %For all remaining inputs...
    for cur=2:m,
        
        ro=rob;         %Set vigilance at its baseline value
        
        if timeslearned(1,A(cur,1,vote))<cap          %Capping the number of elements trained?
            %Calculate coding field signals Tj...
            for j=1:C
                Shyb(j,1)=sum(min(A(cur,2:2*n-1,vote),wijH(:,j,vote)'));   %remember wij = 1 - taoij :)
                Hhyb(j,1)=sum(taoijH(:,j,vote));
                TdistH(j,1)=Shyb(j)+((1-alfa)*Hhyb(j));
                TdistH(j,2)=j;
                if TdistH(j,1)>Tu
                    TdistH(j,3)=TdistH(j,1);          %Stores the value of the Tj's that belong to lambda, others are 0
                else
                    TdistH(j,3)=0;
                end;
                if (((2-alfa)*M)-TdistH(j,1)<0.0000001 & TdistH(j,3)~=0)
                    TdistH(j,4)=1;                   %Indicates poit-box case
                else
                    TdistH(j,4)=0;
                end;
                TdistH(j,5)=Shyb(j,1)/M;             %Vigilance criterium
            end;
            
            %Determine coding field activation
            %CAM rule: denominator of the sum is constant across j's, so
            %it's calculated first
            denom=0;
            for j=1:C                                %For every node in F2
                if TdistH(j,3)~=0                    %if j belongs to lambda
                    denom=denom+((2-alfa)*M-TdistH(j,3))^-p;
                end;
            end;
            
            if sum(TdistH(:,4))~=0
                yH=TdistH(:,4)/sum(TdistH(:,4));     %Point-box case
            else
                yH=1./(((2-alfa)*M-TdistH(:,3)).^p*denom);
                yH(find(TdistH(:,3)==0))=0;
            end;
            
            %Calculate signma_i and check match?
            sicma=zeros(1,2*M);
            for cn=1:C
                sicma=sicma+max(yH(cn)-taojiH(:,cn)',0);
            end;

            %Calculating sigma_k
            
%            predoH=wjkH'*yH;                         %Sum of products rule
            
%            predoH=predoH./classcountH';             %Average of products rule
%            predoH(find(classcountH==0))=0;
            
            for nodo=1:C
                yPred(nodo,:)=yH(nodo)*wjkH(nodo,:);
            end;
            predoH=max(yPred, [], 1)';                %Maximum of products rule
            sigma_k=predoH;
            
            before=predoH;                            %Prediction before enhancement           
            
            %Determine mapping field activation (rule-based inference)
            %Rigido.m describes the set of differential equations
            [T,Y] = ode45(@rigido,[0 15],[0 0 0 0 0 0]);
            filt=Y(length(T),:);
            filt(find(filt<0))=0;
            
            %Per-pixel filtering...
            [prediH,prediValH,yselH]=ppf(filt');
            after=prediH;                           %Prediction after enhancement
            predictoH(cur,:)=prediH;
            
            reseto=after-before';                    % A '1' here means a new element from rule-based inference
            
            %Credit assignment...
            capyselH=yselH(timeslearned(yselH)<cap);        %Capped yselH 
            wjkHtemp=wjkH;
            wjkHtemp(find(wjkHtemp~=1))=0;
            yH=yH.*sum(wjkHtemp(:,capyselH),2);
            %Renormalizing y...
            ToAdd=(1-sum(yH))/length(find(yH~=0));
            yH(find(yH~=0))=yH(find(yH~=0))+ToAdd;  %Re-normalizing corY

            
            %Checking output class prediction...
            if any(yselH==A(cur,1,vote))            %If prediction is correct and satisfies matching...
                
                %increase experience for those nodes
                if length(capyselH)<=4
                    exper(1,capyselH)=exper(1,capyselH)+1;
                end;
                experO(find(exper>experO))=experO(find(exper>experO))+1;        %Experience only increases
                xp=min(maxXP,experO);
                
                %Rule self-organization
                if cur>6*ncat

                    capprediH=prediH;
                    capprediH(find(timeslearned>=cap))=0;     %Prediction for rule learning

                    %fabso(prediH,0,xp);                      %Correct rules confidence is increased
                    fabso(capprediH,0,xp);                    %Rule learning

                    wfab=wrltemp./div;
                    wfab(find(div==0))=0;

                    nwfab=nwrltemp./div;
                    nwfab(find(div==0))=0;

                    mato=wfab-nwfab;
                    mato(find(mato<.5 & mato>-.3))=0;
                    %wfabs(:,:,cur)=mato;

                    wfab(find(wfab<0.5))=0;
                    wfab(find(wfab>=0.5))=1;

                    wfabs(:,:,cur)=mato;

                end;        %end rule self-organization

                %Instar and outstar...
                quanta(cur)=length(find(yH~=0));

                %timeslearned(yselH)=timeslearned(yselH)+1;
                timeslearned(capyselH)=timeslearned(capyselH)+1; 

                sicma=zeros(1,2*M);                     %Sigma_i's for the Outstar first
                for cn=1:C
                    sicma=sicma+max(yH(cn)-taojiH(:,cn)',0);
                end;

                sig=max(sicma-A(cur,2:(2*M+1),vote),0)./sicma;      %First part of the outstar precalculated
                sig(find(sicma==0))=0;

                for cn=1:C
                    updat=yH(cn)-((taoijH(:,cn)'+A(cur,2:(2*M+1),vote)));
                    %updat=yH(cn)-(yH(cn)*(taoijH(:,cn)'+A(cur,2:(2*M+1),vote)));     %Appropriate nodes learn
                    updat(find(updat<0))=0;
                    taoijH(:,cn,vote)=taoijH(:,cn,vote)+Bf*updat';
                    wijH(:,cn,vote)=1-taoijH(:,cn,vote);
                    taojiH(:,cn,vote)=taojiH(:,cn,vote)+sig'.*max(yH(cn)-taojiH(:,cn,vote),0);
                    wjiH(:,cn,vote)=1-taojiH(:,cn,vote);
                end;
                
                %Self-supervised output learninglearning
                wjkH = doSelfSupLearning(wjkH, reseto, xp, yH);
            else
                %Prediction is incorrect or not a good match
                capyselH=yselH(timeslearned(yselH)<cap);        %Capped yselH

                %Decrease experience for those nodes incorrectly predicted
                %exper(1,yselH)=exper(1,yselH)-1;
                exper(1,capyselH)=exper(1,capyselH)-1;    %-(1/(2*length(capyselH)))

                %increase experience for the class that will be learnt
                exper(1,A(cur,1,vote))=exper(1,A(cur,1,vote))+1;
                experO(find(exper>experO))=experO(find(exper>experO))+1;        %Experience only increases
                xp=min(maxXP,experO);
                
                %WTA search...
                tjenordenH=sortrows(TdistH,1);            %Ordering with respect to Tj activation
                
                for j=C:-1:1,                           %Serching for the best node...

                    if j>=2                             %Solving ties (same Tj value)
                        if tjenordenH(j-1,1)==tjenordenH(j,1)
                            tiecount=tiecount+1;
                            ptie=rand;
                            if ptie>.5
                                temp=tjenordenH(j,:);
                                tjenordenH(j,:)=tjenordenH(j-1,:);
                                tjenordenH(j-1,:)=temp;
                                soltie(tiecount)=1;
                            else
                                soltie(tiecount)=0;
                            end;
                        end;
                    end;

                    if tjenordenH(j,1)>alfa*M, 
                        if sum(min(A(cur,2:2*n-1,vote),wijH(:,tjenordenH(j,2),vote)'))/M>=ro  %Matching criterion
                            if wjkH(tjenordenH(j,2),A(cur,1,vote),vote)==1    %Correct prediction
                                %If it gets here, a winner was found!!
                                winner=tjenordenH(j,2);

                                yH(winner)=1;
                                yH(find(yH<1))=0;

                                sicma=zeros(1,2*M);                     %Sigma_i's for the Outstar first
                                for cn=1:C
                                    sicma=sicma+max(yH(cn)-taojiH(:,cn)',0);
                                end;

                                sig=max(sicma-A(cur,2:(2*M+1),vote),0)./sicma;      %First part of the outstar precalculated
                                sig(find(sicma==0))=0;

%                                 tda=yH'*wjiH';
%                                 sig=max(tda-A(cur,2:(2*M+1),vote),0)./tda;      %First part of the outstar precalculated
%                                 sig(find(tda==0))=0;

                                %...And let the learning begin!!
                                for cn=1:C
                                    updat=yH(cn)-((taoijH(:,cn,vote)'+A(cur,2:(2*M+1),vote)));
                                    %updat=yH(cn)-(yH(cn)*(taoijH(:,cn)'+A(cur,2:(2*M+1),vote)));     %Appropriate nodes learn
                                    updat(find(updat<0))=0;
                                    taoijH(:,cn,vote)=taoijH(:,cn,vote)+Bf*updat';
                                    wijH(:,cn,vote)=1-taoijH(:,cn,vote);
                                    taojiH(:,cn,vote)=taojiH(:,cn,vote)+sig'.*max(yH(cn)-taojiH(:,cn,vote),0);
                                    wjiH(:,cn,vote)=1-taojiH(:,cn,vote);
                                end;

                                timeslearned(1,A(cur,1,vote))=timeslearned(1,A(cur,1,vote))+1;

                                %wijH(:,tjenordenH(j,2),vote)=beta*(min(A(cur,2:2*n-1,vote)',wijH(:,tjenordenH(j,2),vote)))+(1-beta)*(wijH(:,tjenordenH(j,2),vote));
                                %taoijH(:,tjenordenH(j,2),vote)=1-wijH(:,tjenordenH(j,2),vote);
                                %classcountD(1,A(cur,1,vote))=classcountD(1,A(cur,1,vote))+1;
                                ok=1;
                                break;
                            else
                                ro=(sum(min(A(cur,2:2*n-1,vote),wijH(:,tjenordenH(j,2),vote)'))/M)+epsilon;   %Match tracking B.9d
                            end;
                        end;
                    end;
                end;
                if ok==0,       %Didn't find an apropriate node... A new node is commited...
                    C=C+1;
                    wijH(:,C,vote)=A(cur,2:2*n-1,vote)';             %Bottom up and...
                    taoijH(:,C,vote)=1-wijH(:,C,vote);
                    wjiH(:,C,vote)=A(cur,2:2*n-1,vote)';             %Top down
                    taojiH(:,C,vote)=1-wjiH(:,C,vote);
                    wjkH(C,A(cur,1,vote),vote)=1;
                    classcountH(1,A(cur,1,vote))=classcountH(1,A(cur,1,vote))+1;
                    timeslearned(1,A(cur,1,vote))=timeslearned(1,A(cur,1,vote))+1;
                    %comnodesH=comnodesH+1;
                else
                    ok=0;
                end;

            end;            %End checking output class prediction
     
        end;    %End capping if
        
    end;    %End cur
    
end;    %End voters

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%Stage 2%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[m,n]=size(te);                             %m is the number of vectors in the test set
te(:,1)=te(:,1)+1;
M=n-1;

te=ordenes(te,1);                        %Create random ordering for the testing set

comptest=[te 1-te];                      %Complement coding
At=[comptest(:,1:n) comptest(:,n+2:2*n)];

At(:,1)=0;

for vote=1:voters,

    predictoH=zeros(m,6);

    for cur=1:m,
        cur
        denom=0;

        for j=1:C,                           %Calculating Tj to determine search order
            TdistH(j,1)=sum(min(At(cur,2:2*n-1),wijH(:,j,vote)'))+(1-alfa)*(M-sum(wijH(:,j,vote)));
            TdistH(j,2)=j;
            if TdistH(j,1)>alfa*M
                TdistH(j,3)=TdistH(j,1);                  %Indicates which belongs to lambda (accord. to Default ARTMAP)
                denom=denom+(1/(M-TdistH(j,1)))^p;
            else
                TdistH(j,3)=0;
            end;

            if TdistH(j,1)==M
                TdistH(j,4)=1;                       %Indicates point-box case (Def. ARTMAP)
            else
                TdistH(j,4)=0;
            end;
        end;

        if sum(TdistH(:,4))==0
            yH=(1./(M-TdistH(:,3))).^p/denom;
        else
            yH=TdistH(:,4)/sum(TdistH(:,4));
        end;

        zigmaH(cur,:)=wjkH'*yH;
        
        for nodo=1:C
            yPred(nodo,:)=yH(nodo)*wjkH(nodo,:);
        end;

        predoH=max(yPred, [], 1)';
        
        for nodo=1:C
            yPred(nodo,:)=yH(nodo)*wjkH(nodo,:);
        end;
        predoH=max(yPred, [], 1)';                %Maximum of products rule
        sigma_k=predoH;

        before=predoH;                            %Prediction before enhancement

        %Determine mapping field activation (rule-based inference)
        %Rigido describes the set of differential equations
        [T,Y] = ode45(@rigido,[0 15],[0 0 0 0 0 0]);
        filt=Y(length(T),:);
        filt(find(filt<0))=0;

        %Per-pixel filtering...
        [prediH,prediValH,yselH]=ppf(filt');
        after=prediH;                           %Prediction after enhancement
        predictoH(cur,:)=prediH;
        reseto=after-before';
        
        %Self-supervised output learninglearning
        wjkH = doSelfSupLearning(wjkH, reseto, xp, yH);
        
        capprediH=prediH;
        capprediH(find(timeslearned>cap))=0;

        %Self-organization continues during testing
        %fabso(prediH,0,xp);
        fabso(capprediH,0,xp);
        %fabso(prediH,ncat,0);
        wfab=wrltemp./div;

        wfab=wrltemp./div;
        wfab(find(div==0))=0;

        nwfab=nwrltemp./div;
        nwfab(find(div==0))=0;

        mato=wfab-nwfab;
        mato(find(mato<.5 & mato>-.3))=0;
        %wfabs(:,:,cur)=wfab;

        wfab(find(wfab<0.5))=0;
        wfab(find(wfab>=0.5))=1;

        if round(cur/20)==cur/20
            wfab;
        end;
        predictoH(cur,1:length(yselH))=yselH;

    end;
end;
